﻿using Microsoft.Extensions.Logging;
using PrecastConcretePlantContracts.BindingModels;
using PrecastConcretePlantContracts.BusinessLogicsContracts;
using PrecastConcretePlantContracts.SearchModels;
using PrecastConcretePlantContracts.StoragesContracts;
using PrecastConcretePlantContracts.ViewModels;
using PrecastConcretePlantDataModels.Enums;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace PrecastConcretePlantBusinessLogic.BusinessLogics
{
    public class OrderLogic : IOrderLogic
    {
        private readonly ILogger _logger;

        private readonly IOrderStorage _orderStorage;

        private readonly IReinforcedStorage _reinforcedStorage;

        private readonly IShopLogic _shopLogic;

        public OrderLogic(ILogger<OrderLogic> logger, IOrderStorage orderStorage, IReinforcedStorage reinforcedStorage, IShopLogic shopLogic)
        {
            _logger = logger;
            _orderStorage = orderStorage;
            _reinforcedStorage = reinforcedStorage;
            _shopLogic = shopLogic;
        }

        public bool CreateOrder(OrderBindingModel model)
        {
            CheckModel(model);
            if (model.Status != OrderStatus.Неизвестен)
            {
                _logger.LogWarning("Invalid order status");
                return false;
            }
            model.Status = OrderStatus.Принят;
            if (_orderStorage.Insert(model) == null)
            {
                _logger.LogWarning("Insert operation failed");
                return false;
            }
            return true;
        }

        public List<OrderViewModel>? ReadList(OrderSearchModel? model)
        {
            _logger.LogInformation("ReadList. OrderId: {Id}.", model?.Id);
            var list = model == null ? _orderStorage.GetFullList() : _orderStorage.GetFilteredList(model);
            if (list == null)
            {
                _logger.LogWarning("ReadList return null list");
                return null;
            }
            _logger.LogInformation("ReadList. Count: {Count}", list.Count);
            return list;
        }

        public bool TakeOrderInWork(OrderBindingModel model)
        {
            return ChangeStatus(model, OrderStatus.Выполняется);
        }

        public bool FinishOrder(OrderBindingModel model)
        {
            return ChangeStatus(model, OrderStatus.Готов);
        }

        public bool DeliveryOrder(OrderBindingModel model)
        {
            return ChangeStatus(model, OrderStatus.Выдан);
        }

        private void CheckModel(OrderBindingModel model, bool withParams = true)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            if (!withParams)
            {
                return;
            }
            if (model.ReinforcedId <= 0)
            {
                throw new ArgumentNullException("Некорректный идентификатор изделия", nameof(model.ReinforcedId));
            }
            if (model.Count <= 0)
            {
                throw new ArgumentNullException("В заказе должно быть хотя бы одно изделие", nameof(model.Count));
            }
            if (model.Sum <= 0)
            {
                throw new ArgumentNullException("Стоимость заказа должна быть больше 0", nameof(model.Sum));
            }
            _logger.LogInformation("Order. Id: {Id}. Sum: {Sum}. ReinforcedId: {ReinforcedId}. ReinforcedCount: {Count}", model.Id, model.Sum,
                model.ReinforcedId, model.Count);
        }

        private bool ChangeStatus(OrderBindingModel model, OrderStatus newStatus)
        {
            CheckModel(model, false);
            var order = _orderStorage.GetElement(new OrderSearchModel { Id = model.Id });
            if (order == null)
            {
                _logger.LogWarning("Change status operation failed. Order not found");
                return false;
            }
            if (order.Status + 1 != newStatus)
            {
                _logger.LogWarning("Change status operation failed. Incorrect new status: {newStatus}. Current status: {currStatus}",
                    newStatus, order.Status);
                return false;
            }
            //наверное тут вопрос 3 тип если изделий нема то и выдать нечего
            if (newStatus == OrderStatus.Выдан)
            {
                var reinforced = _reinforcedStorage.GetElement(new ReinforcedSearchModel() { Id = order.ReinforcedId });
                if (reinforced == null)
                {
                    _logger.LogWarning("Change status operation failed. Reinforced not found");
                    return false;
                }
                if (!_shopLogic.DeliverReinforceds(reinforced, order.Count))
                {
                    _logger.LogWarning("Change status operation failed. Reinforceds delivery operation failed");
                    return false;
                }
            }
            model.Status = newStatus;
            if (model.Status == OrderStatus.Готов)
            {
                model.DateImplement = DateTime.Now;
            }
            else
            {
                model.DateImplement = order.DateImplement;
            }
            if (_orderStorage.Update(model) == null)
            {
                _logger.LogWarning("Change status operation failed");
                return false;
            }
            return true;
        }
    }
}
