﻿using Microsoft.Extensions.Logging;
using PrecastConcretePlantContracts.BindingModels;
using PrecastConcretePlantContracts.BusinessLogicsContracts;
using PrecastConcretePlantContracts.SearchModels;
using PrecastConcretePlantContracts.StoragesContracts;
using PrecastConcretePlantContracts.ViewModels;
using PrecastConcretePlantDataModels.Models;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace PrecastConcretePlantBusinessLogic.BusinessLogics
{
    public class ShopLogic : IShopLogic
    {
        private readonly ILogger _logger;

        private readonly IShopStorage _shopStorage;

        public ShopLogic(ILogger<ShopLogic> logger, IShopStorage shopStorage)
        {
            _logger = logger;
            _shopStorage = shopStorage;
        }

        public List<ShopViewModel>? ReadList(ShopSearchModel? model)
        {
            _logger.LogInformation("ReadList. ShopName: {ShopName}. Id: {Id}", model?.ShopName, model?.Id);
            var list = model == null ? _shopStorage.GetFullList() : _shopStorage.GetFilteredList(model);
            if (list == null)
            {
                _logger.LogWarning("ReadList return null list");
                return null;
            }
            _logger.LogInformation("ReadList. Count: {Count}", list.Count);
            return list;
        }

        public ShopViewModel? ReadElement(ShopSearchModel model)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            _logger.LogInformation("ReadElement.  ShopName: {ShopName}. Id: {Id}", model.ShopName, model.Id);
            var element = _shopStorage.GetElement(model);
            if (element == null)
            {
                _logger.LogWarning("ReadElement element not found");
                return null;
            }
            _logger.LogInformation("ReadElement find. Id: {Id}", element.Id);
            return element;
        }

        public bool Create(ShopBindingModel model)
        {
            CheckModel(model);
            if (_shopStorage.Insert(model) == null)
            {
                _logger.LogWarning("Insert operation failed");
                return false;
            }
            return true;
        }

        public bool Update(ShopBindingModel model)
        {
            CheckModel(model);
            if (_shopStorage.Update(model) == null)
            {
                _logger.LogWarning("Update operation failed");
                return false;
            }
            return true;
        }

        public bool Delete(ShopBindingModel model)
        {
            CheckModel(model, false);
            _logger.LogInformation("Delete. Id: {Id}", model.Id);
            if (_shopStorage.Delete(model) == null)
            {
                _logger.LogWarning("Delete operation failed");
                return false;
            }
            return true;
        }

        public bool ReplenishShop(ShopSearchModel shopModel, IReinforcedModel reinforced, int count)
        {
            if (shopModel == null)
            {
                throw new ArgumentNullException(nameof(shopModel));
            }
            if (reinforced == null)
            {
                throw new ArgumentNullException(nameof(reinforced));
            }
            if (count <= 0)
            {
                throw new ArgumentException("В магазине должен быть хотя бы один товар", nameof(count));
            }
            _logger.LogInformation("ReplenishShop(GetElement). ShopName: {ShopName}. Id: {Id}", shopModel.ShopName, shopModel.Id);
            var shop = _shopStorage.GetElement(shopModel);
            if (shop == null)
            {
                _logger.LogWarning("ReplenishShop(GetElement). Element not found");
                return false;
            }
            if (shop.ReinforcedsMax - shop.ShopReinforceds.Sum(x => x.Value.Item2) < count)
            {
                _logger.LogWarning("ReplenishShop error. No space for new reinforceds");
                return false;
            }
            if (shop.ShopReinforceds.ContainsKey(reinforced.Id))
            {
                var shopR = shop.ShopReinforceds[reinforced.Id];
                shopR.Item2 += count;
                shop.ShopReinforceds[reinforced.Id] = shopR;
                _logger.LogInformation("ReplenishShop. Added {count} '{reinforced}' to '{ShopName}' shop", count, reinforced.ReinforcedName,
                    shop.ShopName);
            }
            else
            {
                shop.ShopReinforceds.Add(reinforced.Id, (reinforced, count));
                _logger.LogInformation("ReplenishShop. Added {count} new '{reinforced}' to '{ShopName}' shop", count, reinforced.ReinforcedName,
                    shop.ShopName);
            }
            if (_shopStorage.Update(new ShopBindingModel()
            {
                Id = shop.Id,
                ShopName = shop.ShopName,
                Address = shop.Address,
                DateOpening = shop.DateOpening,
                ReinforcedsMax = shop.ReinforcedsMax,
                ShopReinforceds = shop.ShopReinforceds,
            }) == null)
            {
                _logger.LogWarning("ReplenishShop. Update operation failed");
                return false;
            }
            return true;
        }

        public bool MakeSale(IReinforcedModel model, int count)
        {
            return _shopStorage.MakeSale(model, count);
        }

        public bool DeliverReinforceds(IReinforcedModel reinforced, int count)
        {
            if (count <= 0)
            {
                _logger.LogWarning("Reinforceds delivery operation failed. Reinforced count <= 0");
                return false;
            }

            var shopList = _shopStorage.GetFullList();
            int shopsCapacity = shopList.Sum(x => x.ReinforcedsMax);
            int currentReinforceds = shopList.Select(x => x.ShopReinforceds.Sum(y => y.Value.Item2)).Sum();
            int freePlaces = shopsCapacity - currentReinforceds;

            if (freePlaces < count)
            {
                _logger.LogWarning("Reinforceds delivery operation failed. No space for new кeinforceds");
                return false;
            }

            foreach (var shop in shopList)
            {
                freePlaces = shop.ReinforcedsMax - shop.ShopReinforceds.Sum(x => x.Value.Item2);
                if (freePlaces == 0)
                {
                    continue;
                }
                if (freePlaces >= count)
                {
                    if (ReplenishShop(new() { Id = shop.Id }, reinforced, count))
                    {
                        count = 0;
                    }
                    else
                    {
                        _logger.LogWarning("Reinforceds delivery operation failed");
                        return false;
                    }
                }
                else
                {
                    if (ReplenishShop(new() { Id = shop.Id }, reinforced, freePlaces))
                    {
                        count -= freePlaces;
                    }
                    else
                    {
                        _logger.LogWarning("Reinforceds delivery operation failed");
                        return false;
                    }
                }
                if (count == 0)
                {
                    return true;
                }
            }
            return false;
        }
        private void CheckModel(ShopBindingModel model, bool withParams = true)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            if (!withParams)
            {
                return;
            }
            if (string.IsNullOrEmpty(model.ShopName))
            {
                throw new ArgumentNullException("Нет названия магазина", nameof(model.ShopName));
            }
            if (string.IsNullOrEmpty(model.Address))
            {
                throw new ArgumentNullException("Нет адреса магазина", nameof(model.Address));
            }
            _logger.LogInformation("Shop. ShopName: {ShopName}. Address: {Address}. Id: {Id}", model.ShopName, model.Address, model.Id);
            var element = _shopStorage.GetElement(new ShopSearchModel
            {
                ShopName = model.ShopName
            });
            if (element != null && element.Id != model.Id)
            {
                throw new InvalidOperationException("Магазин с таким названием уже есть");
            }
        }
    }
}
